;; Generated package description from csv-mode.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "csv-mode" "1.27.0.20260907.5" "Major mode for editing comma/char separated values" '((emacs "27.1") (cl-lib "0.5")) :commit "a693259ddef057eda82421b2a4410b3e35832e91" :url "https://elpa.gnu.org/nongnu/csv-mode.html" :authors '(("Francis J. Wright" . "F.J.Wright@qmul.ac.uk")) :maintainer '(nil . "emacs-devel@gnu.org") :keywords '("convenience"))
